#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "dict.h"

struct dict *
dict_make(void)
{
	struct dict *pt;

	pt = malloc(sizeof *pt);

	pt->nelements = 331;
	pt->props = calloc(pt->nelements, sizeof *pt->props);

	return pt;
}

static struct dict_entry *
dict_entry_make(const char *name, void *value, struct dict_entry *next)
{
	struct dict_entry *p;

	p = malloc(sizeof *p);

	p->name = strdup(name);
	p->value = value;
	p->next = next;

	return p;
}

static unsigned long
hash(const char *str)
{
	unsigned long h = 0;

	while (*str) {
		h = (h << 4) + *(const unsigned char *)str++;
	}

	return h;
}

void
dict_put(struct dict *pt, const char *name, void *value)
{
	int idx;

	idx = hash(name)%pt->nelements;

	pt->props[idx] = dict_entry_make(name, value, pt->props[idx]);
}

void *
dict_get(struct dict *pt, const char *name)
{
	struct dict_entry *p;

	for (p = pt->props[hash(name) % pt->nelements]; p; p = p->next)
		if (!strcmp(p->name, name))
			return p->value;

	return NULL;
}

void
dict_free(struct dict *pt, void(*free_data)(void *))
{
	int i;

	for (i = 0; i < pt->nelements; i++) {
		struct dict_entry *p, *n;

		p = pt->props[i];

		while (p) {
			n = p->next;

			free_data(p->value);
			free(p->name);
			free(p);

			p = n;
		}
	}

	free(pt);
}

void
dict_map(struct dict *pt, void(*mapper)(void *value, void *extra), void *extra)
{
	int i;

	for (i = 0; i < pt->nelements; i++) {
		struct dict_entry *p;

		for (p = pt->props[i]; p; p = p->next)
			mapper(p->value, extra);
	}
}
